using namespace std;
/*************************************************************** 
 * 
 * DalDataSet.cc               & XMM Science Analysis System (SAS) 
 * 
 * Class representing a DAL Data Set.
 * 
 * (c) 1997 ESA - Written by Mark Thomas, CARA
 * 
 * $Id: DalDataSet.cc,v 1.2 2004/03/05 14:43:01 eojero Exp $ 
 * 
 *****************************************************************/ 

#include <STime.h>
#include "Tracer.h"
#include "stringutil.h"
#include "errstr.h"
#include "SimpleScanner.h"

#include "dal_utilities.h"
#include "DalDataSet.h"
#include "DalBlock.h"
#include "DalTable.h"
#include "DalArray.h"


DalDataSet::DalDataSet(
    const string &			name,
    const string &			label,
    AccessMode			openMode,
    DataSetServer::MemoryModel	memoryModel,
    DalDataSetServer *		server )

    :
    DalAttributable( name, label, openMode, 0 ),
    DalChild<DataSetServer>( server ),
    DalChild<DalDataSetServer>( server ),
    _dataSetReaderWriter( 0 ),
    _memoryModel( memoryModel )
{
    PROC
	ENDPROC
	}


DataSet&
DalDataSet::operator=( const DataSet& org )
{
    PROC

	// **********************
	// Delete all attributes.
	// **********************
	const unsigned numberOfAttributes = attributes();

    for( unsigned int i = 0; i < numberOfAttributes; ++i )
    {
	deleteAttribute( 0 );
    } // for

    // ******************
    // Delete all blocks.
    // ******************
    const unsigned int numberOfBlocks = blocks();

    for( unsigned int i = 0; i < numberOfBlocks; ++i )
    {
	delete _blocks[ 0 ];
	_blocks.erase( _blocks.begin() );
    } // for

    // ****************************
    // Copy attributes from source.
    // ****************************
    for( unsigned int i = 0; i < org.attributes(); ++i )
    {
	addAttribute( org.attribute( i ) );
    } // for

    // ************************
    // Copy blocks from source.
    // ************************
    for( unsigned int i = 0; i < org.blocks(); ++i )
    {
	add( org.block( i ) );
    } // for

    RETURN *this;
}


void
DalDataSet::checkValidBlockNumber( unsigned int blockNumber ) const
{
    PROC
	// ***********************************************
	// Check that the requested block number is valid.
	// ***********************************************
	if( blockNumber >= blocks() )
	{
	    errstr << "There is no block " << blockNumber << " in data set " << name() << "; it has " << blocks() << " blocks" << Sas::error("blockNumberNotFound");

	} // if

    ENDPROC
	}


DataSetReaderWriter*
DalDataSet::dataSetReaderWriter() const
{
    PROC
	RETURN _dataSetReaderWriter;
}


string
DalDataSet::parseName( const string & compoundName ) const
{

    SimpleScanner	sc( compoundName, ":%" );
    string		parsedName;

    try
    {
	parsedName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "An error was detected with the name '" << compoundName << "'.  " << e.message() << Sas::error( "invalidCompoundName" );
    }

    if( sc.nextIsCharacter( '%' ) )
    {
	ALTRETURNV( sc.remainder() );
    }
    else if( sc.nextIsCharacter( ':' ) )
    {
	if( parsedName != name() )
	{
	    errstr << "The compound name '" << compoundName << "' is invalid; the qualified name of this object is " << qualifiedName() << Sas::error( "invalidCompoundName" );
	} // if
	sc.skip();

	ALTRETURNV( sc.remainder() );
    }
    else
    {
	ALTRETURNV( compoundName );
    } // if

}


string
DalDataSet::qualifiedName() const
{
    PROC
	RETURN name();
}


//void
//DalDataSet::releaseAll()
//{
//	for(
//		BlockList::iterator i = _blocks.begin();
//		i != _blocks.end();
//		++i )
//	{
//		(*i) -> releaseAll();
//	} // for
//}


void
DalDataSet::checkIfValidBlock(
    const Block*		blk,
    Block::Type		blockType ) const
{
    PROC

	if( blk -> type() != blockType )
	{
	    const string	isThisKindOfBlock = ( blk -> type() == Block::TableT ? "a table" : "an array" );
	    const string	accessAsThisKindOfBlock = ( blockType == Block::TableT ? "a table" : "an array" );

	    errstr << "The block with qualified name " << blk -> qualifiedName() << " is " << isThisKindOfBlock << " and may not be accessed as " << accessAsThisKindOfBlock << "." << Sas::error( "invalidBlockAccess" );
	} // if

    ENDPROC
	}


Table *
DalDataSet::table( unsigned int blockNumber, AccessMode accessMode )
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    checkIfValidBlock( blk, Block::TableT );

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (DalTable *)blk;
}


const Table *
DalDataSet::table( unsigned int blockNumber, AccessMode accessMode ) const
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    checkIfValidBlock( blk, Block::TableT );

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (const DalTable *)blk;
}


Table *
DalDataSet::table( const string & blockName, AccessMode accessMode )
{
    PROC
	RETURN table( blockNumber( blockName ), accessMode );
}


const Table *
DalDataSet::table( const string & blockName, AccessMode accessMode ) const
{
    PROC
	RETURN table( blockNumber( blockName), accessMode );
}


Array *
DalDataSet::array( unsigned int blockNumber, AccessMode accessMode )
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    checkIfValidBlock( blk, Block::ArrayT );

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (DalArray *)blk;
}


const Array *
DalDataSet::array( unsigned int blockNumber, AccessMode accessMode ) const
{

    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    checkIfValidBlock( blk, Block::ArrayT );

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (const DalArray *)blk;
}


Array *
DalDataSet::array( const string & blockName, AccessMode accessMode  )
{
    PROC
	RETURN array( blockNumber( blockName ), accessMode );
}


const Array *
DalDataSet::array( const string & blockName, AccessMode accessMode ) const
{
    PROC
	RETURN array( blockNumber( blockName ), accessMode );
}


Block *
DalDataSet::block( unsigned int blockNumber, AccessMode accessMode )
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (DalBlock *)blk;
}


const Block *
DalDataSet::block( unsigned int blockNumber, AccessMode accessMode ) const
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    blk -> mode( accessMode );
    blk -> alwaysAllowModify( false );

    RETURN (const DalBlock *)blk;
}


Block *
DalDataSet::block( const string & blockName, AccessMode accessMode )
{
    PROC
	RETURN block( blockNumber( blockName ), accessMode );
}


const Block *
DalDataSet::block( const string & blockName, AccessMode accessMode ) const
{
    PROC
	RETURN block( blockNumber( blockName ), accessMode );
}


unsigned int
DalDataSet::blockNumber( const string & blockName ) const
{
    PROC

	errstr << "Attempting to access block with name '"<< blockName <<"' in dataset with name '" << name() << "'." << Sas::message( Msg::Core, Msg::Noisy );

    string	parsedName = parseName( blockName );

    SimpleScanner	sc( parsedName, ":%" );
    unsigned int 	blockNumber = 0;

    if( sc.nextIsCharacter( '%' ) )
    {
	errstr << "The compound name '" << blockName << "' is invalid; the qualified name of this object is '" << qualifiedName() << "'." << Sas::error( "invalidCompoundName" );
    }
    else if( !sc.nextIsCharacter( ':' ) && !sc.eof() )
    {
	try
	{
	    parsedName = sc.match( SimpleToken::String ).asString();
	}
	catch( SimpleScanner::Expected e )
	{
	    errstr << "The compound name '" << blockName << "' is invalid; the qualified name of this object is '" << qualifiedName() << "'." << Sas::error( "invalidCompoundName" );
	} // if

	BlockList::const_iterator i;

	for( i = _blocks.begin(); i != _blocks.end(); ++i, ++blockNumber )
	{
	    if( (*i)->name() == parsedName )
	    {
		break;

	    } // if
	} // for

	if( i == _blocks.end() )
	{
	    if(parsedName == qualifiedName())
		errstr << "I seem to be looking for a block named " << blockName << " in a dataset called " << qualifiedName() << " and I could not find it. Most likely you failed to specify append a ':' to " << qualifiedName() << Sas::error("InvalidCompoundName");
	    else

		errstr << "Could not find block with qualified name '" << qualifiedName() << ":" << parsedName << "' in dataset with name '" << name() << "'" << Sas::error("blockNotFound");

	} // if
    } // if

    RETURN blockNumber;
}


void
DalDataSet::deleteBlock( unsigned int blockNumber )
{

    PROC

	checkIfReadOnly();
    checkValidBlockNumber( blockNumber );

    BlockList::iterator	i = _blocks.begin();

    for( unsigned int j = 0; j < blockNumber; j++ ) ++i;

    // Flush the data
    (*i)->release();
    // **********************************
    // Delete the entry in the data file.
    // **********************************
    if( _dataSetReaderWriter ) _dataSetReaderWriter -> deleteEntry( *(*i) );

    delete *i;

    _blocks.erase( i );

    ENDPROC
	}


void
DalDataSet::deleteBlock( const string & name )
{
    PROC

	checkIfReadOnly();
    deleteBlock( blockNumber( name ) );

    ENDPROC
	}


unsigned int
DalDataSet::blocks() const
{

    PROC
	RETURN _blocks.size();
}


Block::Type
DalDataSet::blockType( unsigned int blockNumber ) const
{
    PROC

	checkValidBlockNumber( blockNumber );

    DalBlock * blk = _blocks[ blockNumber ];

    RETURN blk -> type();

}


Block::Type
DalDataSet::blockType( const string & blockName ) const
{
    PROC

	RETURN blockType( blockNumber( blockName ));
}


bool
DalDataSet::hasBlock( const string & blockName ) const
{
    PROC

	string	parsedName = parseName( blockName );

    SimpleScanner	sc( parsedName, ":%" );

    string	bkName;

    try
    {
	bkName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "The compound name '" << blockName << "' is invalid; the qualified name of this object is '" << qualifiedName() << "'." << Sas::error( "invalidCompoundName" );
    } // if

    bool		found = 0;

    for(
	BlockList::const_iterator i = _blocks.begin();
	i != _blocks.end();
	++i )
    {
	if( bkName == (*i) -> name() )
	{
	    found = 1;

	    break;

	} // if
    } // for

    RETURN found;

}


void
DalDataSet::insert( DalBlock * block, int position )
{
    PROC

	// ***************************************************
	// The intention is to place the new item in the list
	// so that, in the new modified
	// list, it has position Position.
	// ***************************************************

	// ********************************************
	// Check that Position is within a valid range.
	// Position is unsigned and > 0.
	// ********************************************
	unsigned int	pos = (position == -1 ) ? _blocks.size() : position;

    if( pos > _blocks.size() )
    {
	errstr << "Cannot add block with name '" << block -> name() << "' to dataset with name '" << DalDataSet::name() << "' since the specified position " << position << " is invalid; " << DalDataSet::name() << " contains " << blocks() << " blocks." << Sas::error("invalidBlockPosition");
    } // if

    BlockList::iterator	i = _blocks.begin();

    for( unsigned int j = 0; j < pos; ++j ) ++i;

    _blocks.insert( i, block );

    ENDPROC
	}


Table *
DalDataSet::addTable(
    const string &	name,
    unsigned long	rows,
    const string &	label,
    int		        position )
{

    PROC

	checkIfReadOnly();

    string	parsedName = parseName( name );

    SimpleScanner	sc( parsedName, ":%" );

    string blockName;

    try
    {
	blockName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "The compound name '" << name << "' is invalid; the qualified name of this object is '" << qualifiedName() << "'." << Sas::error( "invalidCompoundName" );
    } // if

    if(blockName == "PRIMARY")
	Sas::error("ReservedBlockName", "You cannot add a table with the name PRIMARY because this block name is reserved for the first nameless array in a dataset.");

    if ( hasBlock( blockName ) )
    {
	errstr << "Cannot add table with name '" << blockName << "' because the dataset with name '" << this->name() << "' already contains a block with this name" << Sas::error("blockExists");
    } // if


    DalTable *	newTable = makeTable( blockName, label, rows, this );

    insert( newTable, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    
//    if( _dataSetReaderWriter ) _dataSetReaderWriter -> addEntry( *newTable );
    _dataSetReaderWriter -> addEntry( *newTable );

    RETURN newTable;
}


Array *
DalDataSet::addArray(
    const string &			name,
    DataType			dataType,
    const longVec &			size,
    const string &			units,
    const string &			label,
    int				        position )
{
    PROC

	checkIfReadOnly();

    string	parsedName = parseName( name );

    SimpleScanner	sc( parsedName, ":%" );

    string blockName;

    try
    {
	blockName = sc.match( SimpleToken::String ).asString();
    }
    catch( SimpleScanner::Expected e )
    {
	errstr << "The compound name '" << name << "' is invalid; the qualified name of this object is '" << qualifiedName() << "'." << Sas::error( "invalidCompoundName" );
    } // if

    if ( hasBlock( blockName ) )
	errstr << "Cannot add array with name '" << blockName << "' because the dataset with name '" << this->name() << "' already contains a block with that name" << Sas::error("blockExists");

    DalArray *	newArray = makeArray(
	blockName,
	dataType,
	size,
	units,
	label,
	this );

    insert( newArray, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    if( _dataSetReaderWriter ) _dataSetReaderWriter -> addEntry( *newArray );

    RETURN newArray;
}


Block *
DalDataSet::add(
    const Block *	block,
    const string &	newName,
    int		        position)
{

    PROC
    
	Block * newBlock = 0;
  
    switch( block->type() )
    {
	case Block::TableT :
	{
	    const Table * table = block->parent()->table(block->name());
	    newBlock = add( table, newName, position );
	}
	break;
    
	case Block::ArrayT :
	{
	    const Array *array = block->parent()->array(block->name());
	    newBlock = add( array, newName, position );
	}
	break;
    
	default:
	    errstr << "The block with name '" << block->name() << "' has an invalid type " << int( block->type() ) << Sas::error("internalError");
      
    } // switch
  
    RETURN newBlock;
}


Block *
DalDataSet::add(
    const Table *	block,
    const string &	newName,
    int		        position)
{

    PROC

	checkIfReadOnly();

    const string	blockName = newName.empty() ? block -> name() : newName;

    if ( hasBlock( blockName ))
    {
	errstr << "Cannot add block with name '" << blockName << "' because the set with name '" << name() << "' already contains a block with this name" << Sas::error("blockExists");
    } // if

    DalTable * newBlock = makeTable(
	blockName,
	block -> label(),
	block -> rows(),
	this );

    insert( newBlock, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    if( _dataSetReaderWriter ) _dataSetReaderWriter -> addEntry( *newBlock );

    *newBlock = *block;

    RETURN newBlock;
}


Block *
DalDataSet::add(
    const Array *	block,
    const string &	newName,
    int		        position)
{

    PROC

	checkIfReadOnly();

    const string	blockName = newName.empty() ? block -> name() : newName;

    if ( hasBlock( blockName ))
    {
	errstr << "Cannot add block with name '" << blockName << "' because the  set with name '" << name() << "' already contains a block with this name" << Sas::error("blockExists");
    } // if

    DalArray * newBlock = makeArray(
	blockName,
	block -> dataType(),
	block -> size(),
	block -> units(),
	block -> label(),
	this );

    insert( newBlock, position );

    // ********************************
    // Add an entry into the data file.
    // ********************************
    if( _dataSetReaderWriter ) _dataSetReaderWriter -> addEntry( *newBlock );

    *newBlock = *block;

    RETURN newBlock;
}


void
DalDataSet::rename(
    const string&	newName )
{
    PROC
	DalChild<DalDataSetServer>::parent() -> renameDataSet( name(), newName );
    DalLabelled::rename( newName );
    ENDPROC
	}


void
DalDataSet::release()
{
    PROC
	for( unsigned int i = 0; i < _blocks.size(); ++i )
	{
	    _blocks[ i ] -> release();
	} // for
    ENDPROC
	}


void
DalDataSet::addDebug() const
{
    PROC

	DalDataSetServer*	dataSetServer = const_cast<DalDataSetServer*>( this -> DalChild<DalDataSetServer>::parent() );
    const string	time = STime::currentTime().toString();
    const string	dbg = name() + " "
	+ time + " "
	+ toString( mode() ) + " "
	+ dataSetServer -> client() + " "
	+ toString( memoryModel() ) + " "
	+ dalEnvironment();

    dataSetServer -> debug( this, dbg );

    ENDPROC
	}


DataSetServer::MemoryModel
DalDataSet::memoryModel() const
{
    PROC
	RETURN _memoryModel;
}



