using namespace std;
/*************************************************************** 
* 
* DalDataSetServer.cc         & XMM Science Analysis System (SAS) 
* 
* Class representing the DAL Data Set Server.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: DalDataSetServer.cc,v 1.3 2005/06/06 09:00:10 jhoar Exp $ 
* 
*****************************************************************/ 

#include <string>
#include <fstream>

#include <STime.h>
#include <unistd.h>

#include "Tracer.h"
#include "SimpleScanner.h"
#include "stringutil.h"
#include "DalDataSetServer.h"
#include "MemFitsReaderWriter.h"
#include "HiLowFitsReaderWriter.h"
#include "MemDalReaderWriter.h"
#include "HiLowDalReaderWriter.h"
#include "DalReaderWriter.h"
#include "MemDataSet.h"
#include "HiLowDataSet.h"
#include "DataSetReaderWriter.h"


static DataSetServer *
initialiseDataSetServer()
{
	PROC
	// ************************************
	// Ensure that a singleton is returned.
	// ************************************
	static bool done = false;

	if( !done )
	{
		done = true;
		ALTRETURNV( new DalDataSetServer );
	}
	else
	{	
		ALTRETURNV( dataSetServer );
	} // if
}


DataSetServer * dataSetServer = initialiseDataSetServer();



DalDataSetServer::DalDataSetServer() :
	_defaultModel( getenv( "SAS_MEMORY_MODEL" ) ),
	_clobber( true ),
	_outputFileFormat( DataSetReaderWriter::Fits )
{
	PROC

	char *clobberEnv = getenv("SAS_CLOBBER");
	if (clobberEnv)
		_clobber = atoi(clobberEnv);

	char *formatEnv = getenv("SAS_FORMAT");
	if (formatEnv)
		_outputFileFormat = DataSetReaderWriter::FileType(atoi(formatEnv));

	ENDPROC
}


DalDataSetServer::~DalDataSetServer()
{
	PROC

	// ********************************************************
	// Delete the kept data sets which have not been discarded.
	// ********************************************************
	for(
		RecordIterator i = _records.begin();
		i != _records.end();
		++i )
	{
		//DataSet * dataSet = i -> second;
		DataSetRecord* record = i -> second;

		errstr << "The data set with name '"
			<< record -> dataSet() -> name() 
			<< "' was not discarded"
			<< Sas::message( Msg::Core, Msg::Sparse );
		delete record -> dataSet();
		delete record;
	} // for

  	ENDPROC
}


const string&
DalDataSetServer::client() const
{
	PROC
	RETURN _client;
}


void
DalDataSetServer::client(const string& name)
{
	PROC

	_client = name;

	ENDPROC
}


void
DalDataSetServer::process(const string& processDescription )
{
	PROC

	_process = processDescription;

	ENDPROC
}


const string& 
DalDataSetServer::process() const
{
	PROC
	RETURN _process;
}


void
DalDataSetServer::process(
	const DataSet*		dataSet,
	const string&		processDescription )
{
	PROC

	_processHistory[ dataSet ].push_back( processDescription );

	ENDPROC
}


unsigned int
DalDataSetServer::processes( const DataSet* dataSet ) const
{
	PROC
	ProcessHistory::const_iterator	i = _processHistory.find( dataSet );
	RETURN i == _processHistory.end() ? 0 : (*i).second.size();
}


const string& 
DalDataSetServer::process( const DataSet* dataSet, unsigned int processNumber ) const
{
	PROC
	ProcessHistory::const_iterator	i = _processHistory.find( dataSet );

	if( i == _processHistory.end() )
	{
		errstr << "The dataset with name '"
			<< dataSet->name()
			<< "' does not have any process history" 
			<< Sas::error( "noProcessHistory" );

	} // if

	RETURN (*i).second[ processNumber ];
}


void
DalDataSetServer::debug(
	const DataSet*		dataSet,
	const string&		debugDescription )
{
	PROC

	_debugHistory[ dataSet ].push_back( debugDescription );

	ENDPROC
}


unsigned int
DalDataSetServer::debugs( const DataSet* dataSet ) const
{
	PROC
	DebugHistory::const_iterator	i = _debugHistory.find( dataSet );
	RETURN i == _debugHistory.end() ? 0 : (*i).second.size();
}


const string& 
DalDataSetServer::debug( const DataSet* dataSet, unsigned int debugNumber ) const
{
	PROC
	DebugHistory::const_iterator	i = _debugHistory.find( dataSet );

	if( i == _debugHistory.end() )
	{
		errstr << "The dataset with name '"
			<< dataSet->name()
			<< "' does not have any debug history" 
			<< Sas::error( "noDebugHistory" );

	} // if

	RETURN (*i).second[ debugNumber ];
}


void
DalDataSetServer::creator(
		const DataSet*		dataSet,
		const string&		creator )
{
	PROC
	_creatorOf[ dataSet ] = creator;
	ENDPROC
}


const string&
DalDataSetServer::creator(
		const DataSet*		dataSet ) const
{
	PROC
	CreatorMap::const_iterator	i = _creatorOf.find( dataSet );

	if( i == _creatorOf.end() )
	{
		errstr << "The dataset with name '"
			<< dataSet->name()
			<< "' does not have a creator" 
			<< Sas::error( "creatorOfDataSetNotFound" );

	} // if

	RETURN (*i).second;
}


void
DalDataSetServer::date(
		const DataSet*		dataSet,
		const string		date )
{
	PROC
	_dateOf[ dataSet ] = date;
	ENDPROC
}


const string&
DalDataSetServer::date(
		const DataSet*		dataSet ) const
{
	PROC
	CreatorMap::const_iterator	i = _dateOf.find( dataSet );

	if( i == _dateOf.end() )
	{
		errstr << "The dataset with name '"
			<< dataSet->name()
			<< "' does not have a date" 
			<< Sas::error( "dateOfDataSetNotFound" );

	} // if

	RETURN (*i).second;
}


// *******************************************************
// This function determines the total number of sets which
// are currently in memory. Essentially, it calculates the
// the number of sets in the union of the
// set of _refs and the set of _keeps.
// Note that a set can be in memory because it has been
// opened (referenced) or kept, but we only want to
// count the sets which have been both opened and kept
// once.
// *******************************************************
unsigned int
DalDataSetServer::numberOfSetsInMemory() const
{
	PROC
	unsigned int					sets = 0;

	for( ConstRecordIterator i = _records.begin(); i != _records.end(); ++i )
	{
		if( (*i).second -> inMemory() )
		{
			++sets;
		} // if
	} // for

	RETURN sets;
}


DataSet *
DalDataSetServer::open(
		       const string &		setName,
		       Dal::AccessMode		openMode,
		       MemoryModel		model,
		       DataSetReaderWriter*	readerWriter )
{
  PROC
    
    // FIXME: This should be handled in the parser
    if(setName == "" ||
       setName.find_first_not_of(" ") == std::string::npos)
      Sas::error("EmptySetName", "The dataset name is empty");
  
  const string	cleanName = parseName( setName );
  DataSet*	dataSet;
  MemoryModel	openWithThisModel = DataSetServer::High;
  bool		openedOK = false;
  
  
  errstr << "There are currently "
	 << numberOfSetsInMemory()
	 << " datasets in memory. Request to open a dataset with name '"
	 << cleanName 
	 << "', mode '" 
	 << toString( openMode ) 
	 << "', model '" 
	 << toString( model ) 
	 << Sas::message( Msg::Core, Msg::Sparse );
  
  
  // *******************************************
  // Get the dataset record corresponding to the
  // dataset.
  // *******************************************
  DataSetRecord* record = getRecord( cleanName );
  
  
  // *******************************************************************
  // Increment the reference counter.
  // The close() method decrements the reference counter, and will only
  // close the dataset when the count reaches zero.
  // *******************************************************************
  record -> incRef();
  
  
  // *******************************************************************
  // Set the client name to which the dataset will be associated. The
  // dataset may have been previously opened by another client, and thus
  // the effect is to change the client name to the current client name.
  // The cleanupClient( const string& clientName ) method iterates
  // through all open datasets and closes those (regardless of the
  // reference count) datasets whose client name
  // matches the given client name.
  // *******************************************************************
  record -> client( _client );
  
  
  // *************
  // Set the name.
  // *************
  //getRecord( cleanName ) -> name( cleanName );
  
  
  // **************************************************
  // Check to see if the dataset was previously opened.
  // **************************************************
  if( record -> inMemory() ) {
    dataSet = record -> dataSet();
    
    errstr << "The dataset with name '"
	   << cleanName
	   << "' was previously opened and is already in memory. There have now been "
	   << record -> refs()
	   << " references to this dataset.  It has open mode '"
	   << toString( dataSet -> mode() )
	   << "'."
	   << Sas::message( Msg::Core, Msg::Sparse );
    
  } else {	  
    
    // ************************
    // Choose the memory model.
    // ************************
    switch( model ) {
    case DataSetServer::High :
      openWithThisModel = DataSetServer::High;
      break;
      
      // Low and HighLow are the same
    case DataSetServer::HighLow :
    case DataSetServer::Low :
      openWithThisModel = DataSetServer::HighLow;
      break;
	
    case DataSetServer::UseEnvironment :
	
      switch( _defaultModel.type() )
	{
	case Model::High :
	  openWithThisModel = DataSetServer::High;
	  break;
	    
	case Model::HighLow :	    
	case Model::Low :
	  openWithThisModel = DataSetServer::HighLow;
	  break;
	  
	case Model::Unspecified :
	  Sas::message( Msg::Core, Msg::Sparse,
			"No Memory Model has been specified; Using High Memory Model." );
	  
	  openWithThisModel = DataSetServer::High;
	  break;
	  
	default :
	  Sas::error( "internalError" , "Unknown DAL memory model.");
	  break;
	} // switch
      break;
    default :
      Sas::error( "internalError" , "Unknown DAL memory model.");
      break;
    } // switch
    
    
    // *********************************
    // Choose readerWriter, if required.
    // *********************************
    if( !readerWriter )
      {
	if( openWithThisModel == DataSetServer::High )
	  {
	    if( _outputFileFormat == DataSetReaderWriter::Fits )
	      {
		readerWriter = new MemFitsReaderWriter;
	      }
	    else
	      {
		readerWriter = new MemDalReaderWriter;
	      } // if
	  }
	else
	  {
	    if( _outputFileFormat == DataSetReaderWriter::Fits )
	      {
		readerWriter = new HiLowFitsReaderWriter;
	      }
	    else
	      {
		readerWriter = new HiLowDalReaderWriter;
	      } // if
	  } // if
      } // if	
    
    
    // ***********************************************
    // Now open the dataset using the chosen model.
    // It is necessary to check that the dataset
    // was opened successfully. This is because the
    // constructor for the dataset terminates quietly
    // in the event that the dataset failed to open.
    // This design has been adopted to accommodate the
    // deceit file format which is not supported in
    // the core DAL.
    // Only the High and HighLow models are supported.
    // If the Low memory model is ever implemented
    // the following (as well as the above code)
    // code will have to be modifed to accommodate it.
    // ***********************************************
    if( openWithThisModel == DataSetServer::High )
      {
	MemDataSet * thisDataSet = new MemDataSet(
						  cleanName,
						  openMode,
						  this,
						  readerWriter );
	
	openedOK =
	  openMode == Dal::Temp
	  || thisDataSet -> dataSetReaderWriter() != 0;
	dataSet = thisDataSet;
      }
    else
      {
	HiLowDataSet * 	thisDataSet = new HiLowDataSet(
						       cleanName,
						       openMode,
						       this,
						       readerWriter );
	
	openedOK = thisDataSet -> dataSetReaderWriter() != 0;
	dataSet = thisDataSet;
      } // if
    
    // *******************************************
    // Check that dataset was successfully opened.
    // *******************************************
    if( !openedOK )
      {
	errstr << "The dataset with name '"
	       << cleanName
	       << "' could not be opened, probably because the format was not recognised." 
	       << Sas::error( "couldNotOpenDataSet" );
	
      } // if
    
    
    
    // *************************************************************
    // Now write a message that the dataset was opened successfully.
    // *************************************************************
    errstr << "The dataset with name '" 
	   << cleanName 
	   << "' was opened successfully with memory model '" 
	   << toString( openWithThisModel ) 
	   << "'." 
	   << Sas::message( Msg::Core, Msg::Sparse );
    
    // ******************************************
    // Check to see if the dataset is to be kept.
    // ******************************************
    if( record -> mustKeep() )
      {
	errstr << "The dataset with name '"
	       << cleanName 
	       << "' will be kept in memory." 
	       << Sas::message( Msg::Core, Msg::Sparse );
	
      } // if
    
    // ****************
    // Set the dataset.
    // ****************
    record -> dataSet( dataSet );
  }
    
  if(openMode == Dal::Modify)
    dataSet->addHistory("Modified by "+ _client + " at " +
			STime::currentTime().toString(0));
  if(openMode == Dal::Create)
    dataSet->addHistory("Created by "+ _client + " at " +
			STime::currentTime().toString(0));


  // FIXME: this has nothing to do with the DAL
  // Add LONGSTRN attribute so that fverify does not complain.
  if(!dataSet->hasAttribute("LONGSTRN") && openMode != Dal::Read)
      *(dataSet->addAttribute("LONGSTRN")) = "OGIP 1.0";
  

  RETURN dataSet;
}


DataSet*
DalDataSetServer::clone(
		const string&		from,
		const string&		to,
		Dal::AccessMode		openMode,
		MemoryModel		model,
		DataSetReaderWriter*	readerWriter )
{
	PROC

	DataSet* clonedSet = 0;
	DataSet* sourceSet = 0;

	switch( openMode )
	{
		case Dal::Temp :
			clonedSet = DalDataSetServer::open( to, openMode, model, readerWriter );
			sourceSet = DalDataSetServer::open( from, Dal::Read, DataSetServer::Low, 0 );
			*clonedSet = *sourceSet;
			DalDataSetServer::close( sourceSet );

			break;

		case Dal::Modify :
			DalDataSetServer::copy( from, to );
			clonedSet = DalDataSetServer::open( to, openMode, model, readerWriter );

			break;

		case Dal::Read :
			errstr << "Clone/Read combination is invalid"
				<< Sas::error( "invalidCloneReadCombination" );

			break;

		case Dal::Create :
			errstr << "Clone/Create combination is invalid" 
				<< Sas::error( "invalidCloneCreateCombination" );

			break;

		default :
			errstr << "Unrecognised access mode: " 
				<< int(openMode) 
				<< Sas::error( "internalError" );

	} // switch

	RETURN clonedSet;
}


void
DalDataSetServer::remove( const string& setName )
{
	PROC

	const string	cleanName = parseName( setName );

	errstr << "Attempting to remove the dataset with name '" 
		<< cleanName 
		<< "'" 
		<< Sas::message( Msg::Core, Msg::Sparse );

	if( !exists( cleanName ) )
	{
		errstr << "The dataset with name '" 
			<< cleanName 
			<< "' cound not be deleted; either it does not exist or it is in an unrecognised format" 
			<< Sas::error( "deleteDataSetFailed" );

	} // if

	getRecord( cleanName ) -> remove();

	ENDPROC
}


void
DalDataSetServer::cleanupClient( const string& clientName )
{
	PROC
	for( RecordIterator i = _records.begin(); i != _records.end(); ++i )
	{
		DataSetRecord* record = (*i).second;

		if(
			record -> client() == clientName 
			&& 
			record -> inMemory() )
		{
			// ********************************************
			// The following will delete the dataset. Later
			// it may be necessary to chnage this
			// if the discard() method becomes conditional.
			// ********************************************
			record -> discard();

			//while( record -> refs() > 1 ) record -> decRefs();
			//close( record -> dataSet() );
			//remove( (*i).second -> name() );

		} // if
	} // for
	ENDPROC
}


void
DalDataSetServer::clobberDataSet( const string& setName )
{
	PROC

	const string	cleanName = parseName( setName );
	const bool	setExistsOnDisk = getRecord( cleanName ) -> onDisk();

	// ****************************************************
	// Check to see if the dataset may be clobbered.
	// If it may, then attempt to clobber it, if necessary.
	// ****************************************************
	if( setExistsOnDisk && !clobber() )
	{
		errstr << "The dataset with name '"
			<< cleanName 
			<< "' could not be clobbered because it already exists and clobber is not set." 
			<< Sas::error( "clobberFailed" );

	}
	else if( setExistsOnDisk )
	{
		// ====================
		// Clobber the dataset.
		// ====================
		status = unlink( cleanName.c_str() );	// UNIX delete file.

		if( status != 0 )
		{
			errstr << "The dataset with name '"
				<< cleanName 
				<< "' exists but could could not be clobbered "
				<< "(returned with status " << status << ")"  
				<< Sas::error("clobberFailed");

		} // if
	}
	else if( clobber() )
	{
		// ===================================
		// Clobber the file , if there is one.
		// ===================================
		unlink( cleanName.c_str() );	// UNIX delete file.
	} // if


	ENDPROC
}


void
DalDataSetServer::close( DataSet * dataSet )
{
	PROC

	const string	setName = dataSet -> name();

	// **************
	// Debug message.
	// **************
	errstr << "Request to close the dataset with name '"
		<< setName 
		<< "'.  There are currently " 
		<< numberOfSetsInMemory() 
		<< " datasets in memory" 
		<< Sas::message( Msg::Core, Msg::Sparse );

	// ********************************
	// Decrement the reference counter.
	// ********************************
	//_records[ setName ].decRef();

	// ******************************
	// Add CREATOR & DATE attributes.
	// Add history records.
	// ******************************

//  	string time = STime::currentTime().toString(0);

//  	switch ( dataSet -> mode() )
//  	{

//  		case Dal::Create:
//  		case Dal::Temp:
//  			//*dataSet->addAttribute("CREATOR","name of code that generated this file") =  _client;
//  			//*dataSet->addAttribute("DATE","creation date") = time;
//  			dataSet->addHistory("Created by "+_client+" at "+time);
//  			//dataSet->addHistory("Process: "+_process);
//  			break;

//  		case Dal::Modify:
//  			dataSet->addHistory("Modified by "+_client+" at "+time);
//  			//dataSet->addHistory("Process: "+_process);
//  			break;

//  		case Dal::Read:
//  			break;

//  		default:
//  			errstr << "invalid open mode " 
//  				<< int( dataSet -> mode() ) 
//  				<< " for dataset with name '" 
//  				<< dataSet -> name() 
//  				<< "'." 
//  				<< Sas::error("internalError");

//  	} // switch

	getRecord( setName ) -> close();

	// ***************************************************
	// Remove entry (if present) from _processHistory map.
	// ***************************************************
	_processHistory.erase( dataSet );

	// *************************************************
	// Remove entry (if present) from _debugHistory map.
	// *************************************************
	_debugHistory.erase( dataSet );

	ENDPROC
}


bool
DalDataSetServer::exists( const string & setName ) const
{
	PROC
	const string	cleanName = parseName( setName );
	RETURN getRecord( cleanName ) -> exists();
	//ConstRecordIterator ri = _records.find( cleanName );
	//if( ri == _records.end() )
	//{
	//	getRecord( cleanName );
	//	RETURN exists( cleanName );
		//RETURN false;
	//}
	//else
	//{
	//	RETURN (*ri).second -> exists();
	//} // if
	//RETURN getRecord( cleanName ) -> exists();
	//RETURN (*(_records.find( cleanName ))).second.exists();
}


void
DalDataSetServer::copy(
	const string&		from,
	const string&		to )
{
	PROC

	const string	f = parseName( from );
	const string	t = parseName( to );

	// ***************************************
	// Check to see if the source file exists.
	// ***************************************
	if( !dataSetServer -> exists( from ) )
	{
		errstr << "It was not possible to copy the dataset with name '" 
			<< f 
			<< "' to '" 
			<< t 
			<< "' because '" 
			<< f 
			<< "' does not exist." 
			<< Sas::error( "copyDataSetFailed" );

	} // if


	// **************************************
	// Use the system function to execute cp.
	// **************************************
	string cmd = "cp "+f+" "+t;

	if( system(cmd.c_str()) )
	{
		errstr << "It was not possible to copy the dataset with name '" 
			<< f 
			<< "' to '" 
			<< t 
			<< "'" 
			<< Sas::error( "copyDataSetFailed" );

	} // if

	ENDPROC
}


void
DalDataSetServer::printOn(ostream&) const
{
	PROC
	ENDPROC
}


void
DalDataSetServer::clobber( bool onoff )
{
	PROC
	_clobber = onoff;
	ENDPROC
}


bool
DalDataSetServer::clobber() const
{
	PROC
	RETURN _clobber;
}


string
DalDataSetServer::parseName( const string & compoundName ) const
{
	PROC
	SimpleScanner	sc( compoundName, ":%" );
	string		parsedName;

	try
	{
		parsedName = sc.match( SimpleToken::String ).asString();
	}
	catch( SimpleScanner::Expected e )
	{
		errstr << "An error was detected with the name '" 
			<< compoundName 
			<< "'.  " 
			<< e.message() 
			<< Sas::error( "invalidCompoundName" );

	}
	RETURN parsedName;
} 


void
DalDataSetServer::keep(
		const string&		setName )
{
	PROC
	const string	cleanName = parseName( setName );
	getRecord( cleanName ) -> keep();
	ENDPROC
}


void
DalDataSetServer::discard(
		const string&		setName )
{
	PROC
	const string	cleanName = parseName( setName );
	getRecord( cleanName ) -> discard();
	ENDPROC
}


void
DalDataSetServer::renameDataSet(
		const string& 		oldName,
		const string&		newName )
{
	PROC
	//getRecord(  oldName ].name( newName );
	//DataSetRecord* record = getRecord( oldName );
	RecordIterator i = _records.find( oldName );
	DataSetRecord* oldRecord = (*i).second;
	_records.erase( i );
	getRecord( newName ) = oldRecord;
	ENDPROC
}


DalDataSetServer::DataSetRecord*&
DalDataSetServer::getRecord( const string& name ) const
{
	PROC
	if( _records.find( name ) == _records.end() )
	{
		DataSetRecord* newRecord = new DataSetRecord( name, _client );
		_records[ name ] = newRecord;
	} // if
	RETURN (*_records.find( name )).second;
}


// ****************************************************************
// DalDataSetServer::DataSetRecord methods are defined beyond here.
// ****************************************************************
DalDataSetServer::DataSetRecord::DataSetRecord(
		const string& name,
		const string& client )
{
	PROC
	_dataSet = 0;
	_keep = false;
	_refs = 0;
	_client = client;
	_name = name;
	ENDPROC
}


void
DalDataSetServer::DataSetRecord::dataSet( DataSet* dataSet )
{
	PROC 
	_dataSet = dataSet;
	ENDPROC
}


DataSet*
DalDataSetServer::DataSetRecord::dataSet()
{
	PROC
	RETURN _dataSet;
}


void
DalDataSetServer::DataSetRecord::incRef()
{
	PROC
	_refs++;
	ENDPROC
}


void
DalDataSetServer::DataSetRecord::decRefs()
{
	PROC
	_refs--;
	ENDPROC
}


unsigned int 
DalDataSetServer::DataSetRecord::refs()
{
	PROC RETURN _refs;
}


bool
DalDataSetServer::DataSetRecord::mustKeep() const
{
	PROC
	RETURN _keep;
}


void 
DalDataSetServer::DataSetRecord::keep()
{
	PROC
	_keep = true;
	ENDPROC
}


void
DalDataSetServer::DataSetRecord::discard()
{
	PROC
	if( _dataSet ) {
		delete _dataSet;
		_dataSet = 0;
	} // if
	_keep = false;
	ENDPROC
}


bool
DalDataSetServer::DataSetRecord::inMemory() const
{
	PROC
	RETURN _dataSet != 0;
}


bool
DalDataSetServer::DataSetRecord::onDisk() const
{
	PROC
	if(
		FitsReaderWriter::exists( _name )
		||
		DalReaderWriter::exists( _name ) )
	{

		ALTRETURNV( true );

	} // if
	RETURN false;
}


bool
DalDataSetServer::DataSetRecord::exists() const
{
	PROC
	RETURN inMemory() || onDisk();
}


void
DalDataSetServer::DataSetRecord::remove()
{
	PROC
	if( inMemory() )
	{
		delete _dataSet;

		errstr << "The dataset with name '" 
			<< _name 
			<< "' was removed successfully from memory" 
			<< Sas::message( Msg::Core, Msg::Sparse );

		_dataSet = 0;
		_keep = false;
		_refs = 0;
		_client = "";
		//_name = "";

	} // if

	if( onDisk() )
	{
		// **********************************
		// Now remove the file from the disk.
		// **********************************
		int status = unlink( _name.c_str() );	// UNIX delete file.

		if( status )
		{
			errstr << "The dataset with name '"
				<< _name 
				<< "' could not be deleted from disk." 
				<< Sas::error( "deleteDatasetFailed" );

		} // if

		errstr << "The dataset with name '" 
			<< _name 
			<< "' was removed successfully from disk" 
			<< Sas::message( Msg::Core, Msg::Sparse );

	} // if
	ENDPROC
}


void
DalDataSetServer::DataSetRecord::close()
{
	PROC

	decRefs();

	// ***********************************************************
	// If the dataset is not to be kept in memory, then delete it.
	// ***********************************************************
	const bool	keep = _keep || _refs;

	if( keep )
	{
		errstr << "The dataset with name '" <<
			_name << "' will not be closed, and will be kept in memory. There are still " 
			<< _refs 
			<< " references to this dataset." 
			<< Sas::message( Msg::Core, Msg::Sparse );

	}
	else
	{
		delete _dataSet;

		// *************************************************************
		// Now write a message that the dataset was closed successfully.
		// *************************************************************
		errstr << "The dataset with name '" 
			<< _name 
			<< "' was closed successfully." 
			<< Sas::message( Msg::Core, Msg::Sparse );

		// ****************************************
		// Reset the record to its initial default.
		// ****************************************
		_dataSet = 0;
		//_keep = false;
		// _refs = 0;
		_client = "";
		//_name = "";
	} // if
	ENDPROC
}


/*
void
DalDataSetServer::DataSetRecord::name( const string& name )
{
	PROC
	_name = name;
	ENDPROC
}
*/


const string&
DalDataSetServer::DataSetRecord::name() const
{
	PROC
	RETURN _name;
}



void
DalDataSetServer::DataSetRecord::client( const string& client )
{
	PROC
	_client = client;
	ENDPROC
}


const string&
DalDataSetServer::DataSetRecord::client() const
{
	PROC
	RETURN _client;
}





