/*************************************************************** 
* 
* Data.h                     & XMM Science Analysis System (SAS) 
* 
* Class representing data object.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: Data.h,v 1.2 2004/03/05 14:43:02 eojero Exp $ 
* 
*****************************************************************/ 
#ifndef DATA_H
#define DATA_H


#include <string>
#include <vector>


#include "MemChild.h"


template<class T>
class Data : public MemChild<T>
{
	public :

		enum DataType { Bool = 1, Int8, Int16, Int32, Real32, Real64, DString };

		// Constructors
		Data(
			DataType			dataType,
			unsigned long			rows,
			const std::vector<unsigned long> &	size,
			const T *			parent )

			:	_dataType( dataType ),
				_rows( rows ),
				_size( size ),
				MemChild<T>( parent )
		{
			PROC

			switch( dataType )
			{
				case Data::Int8 :
					_data = new int8[ elements() ];
					break;
				case Data::Int16 :
					_data = new int16[ elements() ];
					break;
				case Data::Int32 :
					_data = new int32[ elements() ];
					break;
				case Data::Real32 :
					_data = new real32[ elements() ];
					break;
				case Data::Real64 :
					_data = new real64[ elements() ];
					break;
				case Data::Bool :
					_data = new bool[ elements() ];
					break;
				case Data::DString :
					_data = new std::string[ elements() ];
					break;
				default :
					break;
			} // switch

			ENDPROC
		}


		~Data();

		int8 * int8Data() const;
		int16 * int16Data() const;
		int32 * int32Data() const;
		real32 * real32Data() const;
		real64 * real64Data() const;
		bool * boolData() const;
		std::string * stringData() const;
		unsigned long elements() const;
		unsigned int dataSize() const;


	private :

		DataType 		_dataType;
		unsigned long		_rows;
		std::vector<unsigned long>	_size;
		void *			_data;

};


// ******************************
// Inlined functions beyond here.
// ******************************

#include "Tracer.h"


template<class T>
inline Data<T>::~Data()
{
	PROC
	delete _data;
	ENDPROC
}


template<class T>
inline int8 *
Data<T>::int8Data() const
{
	PROC
	RETURN (int8 *)_data;

}


template<class T>
inline int16 *
Data<T>::int16Data() const
{
	PROC
	RETURN (int16 *)_data;

}


template<class T>
inline int32 *
Data<T>::int32Data() const
{
	PROC
	RETURN (int32 *)_data;

}


template<class T>
inline real32 *
Data<T>::real32Data() const
{
	PROC
	RETURN (real32 *)_data;

}


template<class T>
inline real64 *
Data<T>::real64Data() const
{
	PROC
	RETURN (real64 *)_data;

}


template<class T>
inline bool *
Data<T>::boolData() const
{
	PROC
	RETURN (bool *)_data;

}


template<class T>
inline std::string *
Data<T>::stringData() const
{
	PROC
	RETURN (std::string *)_data;

}


template<class T>
inline unsigned long
Data<T>::elements() const
{
	PROC

	unsigned long	numberOfElements = _rows;

	for( unsigned int i = 0; i < _size.size(); ++i )
	{
		numberOfElements *= _size[ i ];
	} // for

	RETURN numberOfElements;
}


template<class T>
inline unsigned int
Data<T>::dataSize() const
{
	unsigned int size;

	switch( _dataType )
	{
		case Data::Int8 :
			size = sizeof( int8 );
			break;
		case Data::Int16 :
			size = sizeof( int16 );
			break;
		case Data::Int32 :
			size = sizeof( int32 );
			break;
		case Data::Real32 :
			size = sizeof( real32 );
			break;
		case Data::Real64 :
			size = sizeof( real64 );
			break;
		case Data::Bool :
			size = sizeof( bool );
			break;
		case Data::DString :
			size = sizeof( std::string * );
			break;
		default :
			break;
	} // switch

	RETURN size;

}



#endif

