using namespace std;
/*************************************************************** 
* 
* MemDataSet.cc              & XMM Science Analysis System (SAS) 
* 
* Class representing a DAL data set.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: MemDataSet.cc,v 1.2 2004/03/05 14:43:03 eojero Exp $ 
* 
*****************************************************************/ 


#include <fstream>			// C++

#include <errstr.h>			// SAS
#include <Tracer.h>

#include "Dal.h"			// DAL
#include "MemDataSet.h"
#include "MemTable.h"
#include "MemArray.h"


void
MemDataSet::readmodify()
{
	PROC

	errstr << "Reading dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );

	alwaysAllowModify( true );

	// **************
	// Read the file.
	// **************
	_dataSetReaderWriter -> read( this );

	alwaysAllowModify( false );

	ENDPROC
}


void
MemDataSet::create()
{
	PROC
	errstr << "Creating dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );
	ENDPROC
}


void
MemDataSet::temp()
{
	PROC
	errstr << "Creating temporary dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );
	ENDPROC
}


MemDataSet::MemDataSet(
			const string &			setName,
			AccessMode			openMode,
			DalDataSetServer *              server,
			DataSetReaderWriter*		readerWriter )
	 :
	 	DalDataSet( setName, "", openMode, DataSetServer::High, server )
{

	PROC

	errstr << "Accessing dataset with name '" << name() << "'" << Sas::message( Msg::Core, Msg::Verbose );

	_dataSetReaderWriter = readerWriter;

	addDebug();

	switch( mode() )
	{
		case Read :
		case Modify :

			readmodify();

			break;


		case Create :

			create();

			break;

		case Temp :

			temp();

			break;

		default :

			errstr << "The DAL could not create data set '" << name() << "' since its access mode "<< int( openMode ) <<" is invalid" << Sas::error("internalError");

			break;

	} // switch


	ENDPROC

}


MemDataSet::MemDataSet( const MemDataSet & dataSet )
	:
  Dal(),
  Labelled(),
  Attributable(),
  Child<DataSetServer>(),
  DataSet(),
  Child<DalDataSetServer>(),
  DalDataSet( name(), label(), mode(), memoryModel(), DalChild<DalDataSetServer>::parent() )
{
	PROC
	_blocks = dataSet._blocks;

	//switch( _dataSetReaderWriter -> fileType() )
	//{
	//	case DataSetServer::Fits :
	//		_dataSetReaderWriter = new MemFitsReaderWriter( this );
	//		break;

	//	case DataSetServer::Dal :
	//		_dataSetReaderWriter = new MemDalReaderWriter( this );
	//		break;

	//	default :
	//		errstr << "The file type (with value " << int( _dataSetReaderWriter -> fileType() ) << ") is not recognised" << Sas::error( "InternalError" );
	//		break;

	//} // switch

	ENDPROC
}


DalTable *
MemDataSet::makeTable(
			const string &		name,
			const string &		label,
			unsigned long		rows,
			DalDataSet *		parent )
{
	PROC
	RETURN new MemTable( name, rows, label, parent );
}


DalArray *
MemDataSet::makeArray(
			const string &			name,
			DataType			dataType,
			const longVec&			size,
			const string &			units,
			const string &			label,
			DalDataSet *			parent )
{
	PROC
	RETURN new MemArray( name, dataType, size, units, label, parent );
}


//DalDataSet &
//MemDataSet::operator=( const DalDataSet & /*dataSet*/ )
//{
//	PROC
	//(Attributable &)*this = dataSet;
	//(DalDataSet &)*this = dataSet;
	//_blocks = dataSet._blocks;
	//_mode = dataSet.mode();

	// TODO: IMPLEMENT
	//switch( dataSet._fileReaderWriter -> fileType() )
	//{
	//	case DataSetServer::Fits :

	//		_fileReaderWriter = new MemFitsFileReaderWriter( this );

	//		break;

	//	case DataSetServer::Dal :

	//		_fileReaderWriter = new MemDalFileReaderWriter( this );

	//		break;

	//	default :

	//		errstr << "The file type (with value " << int( dataSet._fileReaderWriter -> fileType() ) << ") is not recognised" << Sas::error( "InternalError" );

	//		break;

	//} // switch

//	RETURN *this;
//}


MemDataSet::~MemDataSet()
{
  //PROC

	if( _dataSetReaderWriter )
	{
		switch( mode() )
		{
			case Create :

				_dataSetReaderWriter -> clobber( DalChild<DataSetServer>::parent() -> clobber() );

				// ***************
				// Write the file.
				// ***************
				_dataSetReaderWriter -> write( this );

				break;

			case Modify :

				_dataSetReaderWriter -> clobber( true );

				// ***************
				// Write the file.
				// ***************
				_dataSetReaderWriter -> write( this );

				break;

			case Read :
			case Temp :

				break;

			default :

				errstr << "The DAL could not close data set '" << name() << "' since its access mode "<< int( mode() ) <<" is invalid" << Sas::error("internalError");

				break;

		} // switch

		// **************************
		// Delete the dataset blocks.
		// **************************
		for(
			BlockList::iterator i = _blocks.begin();
			i != _blocks.end();
			++i )
		{
		  DalBlock* block=(*i);
		  delete block;
		  (*i)=0;
		} // for

		delete _dataSetReaderWriter;

	} // if

	ENDPROC

}


void
MemDataSet::printOn( ostream & os ) const
{
	PROC

	// ************************
	// Write the Data Set Name.
	// ************************
	os << "DATA SET: Name->" << name() << endl;

	// **************************
	// Write the Data Set Header.
	// **************************
	DalAttributable::printOn( os );


	// **************************
	// Write the Data Set Blocks.
	// **************************
	os << "There are " << _blocks.size() << " blocks" << endl;

	unsigned int no = 1;

	for(
		BlockList::const_iterator i = _blocks.begin();
		i != _blocks.end();
		++i, ++no )
	{
		os << "[" << no << "]";

		(*i) -> printOn( os );

	} // for

	ENDPROC
}


